import os

DATABASES = {
    'default': {
        'ENGINE': os.environ.get('DB_ENGINE', 'django.db.backends.postgresql_psycopg2'),
        'NAME': 'gadjo-test-%s' % os.environ.get('BRANCH_NAME', '').replace('/', '-')[:45],
    }
}

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [],
            'builtins': [
                'gadjo.templatetags.gadjo',
            ],
        },
    },
]


DEBUG = True
USE_TZ = True
INSTALLED_APPS = [
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sites',
    'gadjo',
]
STATIC_URL = '/static/'
SITE_ID = 1
MIDDLEWARE_CLASSES = ()
LOGGING = {}
SECRET_KEY = 'yay'



from django import forms
from django.template import Context, Template
from django.test.client import RequestFactory
from pyquery import PyQuery

from gadjo.forms.widgets import MultiSelectWidget


def test_multiselect_widget():
    class ExampleForm(forms.Form):
        choices = forms.MultipleChoiceField(
            label='choices', choices=[('a', 'Aa'), ('b', 'Bb'), ('c', 'Cc')], widget=MultiSelectWidget
        )

    request = RequestFactory().get('/')
    t = Template('{{ form|with_template }}')
    ctx = Context({'request': request, 'form': ExampleForm()})
    rendered = t.render(ctx)
    assert len(PyQuery(rendered).find('select')) == 1
    assert PyQuery(rendered).find('.gadjo-multi-select-widget--button-add')

    request = RequestFactory().get('/?choices=a&choices=b')
    t = Template('{{ form|with_template }}')
    ctx = Context({'request': request, 'form': ExampleForm(data=request.GET)})
    rendered = t.render(ctx)
    assert len(PyQuery(rendered).find('select')) == 2
    assert PyQuery(rendered).find('option[selected]').text() == 'Aa Bb'
    assert ctx['form'].cleaned_data == {'choices': ['a', 'b']}


def test_multiselect_widget_omitted_from_data():
    class ExampleForm(forms.Form):
        choices = forms.MultipleChoiceField(
            label='choices', choices=[('a', 'Aa'), ('b', 'Bb'), ('c', 'Cc')], widget=MultiSelectWidget
        )

    form = ExampleForm(data={})
    assert (
        form.fields['choices'].widget.value_omitted_from_data(
            form.data, form.files, form.add_prefix('choices')
        )
        is True
    )

    form = ExampleForm(data={'choices': 'a'})
    assert (
        form.fields['choices'].widget.value_omitted_from_data(
            form.data, form.files, form.add_prefix('choices')
        )
        is False
    )


def test_container_attrs():
    class ExampleForm(forms.Form):
        foobar = forms.CharField(label='foobar')

        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.fields['foobar'].widget.container_attrs = {'data-xxx': 'blah', 'data-yyy': '$prefixblah'}

    request = RequestFactory().get('/')
    t = Template('{{ form|with_template }}')
    ctx = Context({'request': request, 'form': ExampleForm()})
    rendered = t.render(ctx)
    assert PyQuery(rendered).attr['data-xxx'] == 'blah'
    assert PyQuery(rendered).attr['data-yyy'] == 'blah'

    request = RequestFactory().get('/')
    t = Template('{{ form|with_template }}')
    ctx = Context({'request': request, 'form': ExampleForm(prefix='c123')})
    rendered = t.render(ctx)
    assert PyQuery(rendered).attr['data-xxx'] == 'blah'
    assert PyQuery(rendered).attr['data-yyy'] == 'c123-blah'



from django import forms


class MultiSelectWidget(forms.MultiWidget):
    template_name = 'gadjo/widgets/multiselectwidget.html'

    class Media:
        js = ('js/gadjo.multiselectwidget.js',)
        css = {'all': ('css/gadjo.multiselectwidget.css',)}

    def __init__(self, attrs=None, form_field=None, max_choices=None):
        self.attrs = attrs
        self.form_field = form_field or forms.Select
        self.max_choices = max_choices
        widgets = [self.form_field(attrs=attrs)]
        super().__init__(widgets, attrs)

    def get_context(self, name, value, attrs):
        if not isinstance(value, list):
            value = [value]

        self.widgets = []
        for _ in range(max(len(value), 1)):
            self.widgets.append(self.form_field(attrs=self.attrs, choices=self.choices))

        # all subwidgets must have the same name
        self.widgets_names = [''] * len(self.widgets)

        context = super().get_context(name, value, attrs)
        context['widget']['max_choices'] = self.max_choices

        return context

    def decompress(self, value):
        return value or []

    def value_from_datadict(self, data, files, name):
        values = [x for x in data.getlist(name) if x]

        # remove duplicates while keeping order
        return list(dict.fromkeys(values))

    def id_for_label(self, id_):
        return id_

    def value_omitted_from_data(self, *args, **kwargs):
        return super(forms.MultiWidget, self).value_omitted_from_data(*args, **kwargs)



import html
import urllib

from django import forms
from django.template import Context, Template
from django.test.client import RequestFactory
from pyquery import PyQuery


def test_start_timestamp():
    t = Template('{% start_timestamp %}')
    assert t.render(Context())


def test_querystring():
    t = Template('{% querystring "name"="Ayers" "age"=20 %}')
    ctx = Context({'request': RequestFactory().get('/')})
    assert urllib.parse.parse_qs(urllib.parse.urlparse(html.unescape(t.render(ctx))).query) == {
        'age': ['20'],
        'name': ['Ayers'],
    }
    ctx = Context({'request': RequestFactory().get('/?age=10')})
    assert urllib.parse.parse_qs(urllib.parse.urlparse(html.unescape(t.render(ctx))).query) == {
        'age': ['20'],
        'name': ['Ayers'],
    }

    t = Template('{% querystring "name"="Ayers" without "gender" %}')
    ctx = Context({'request': RequestFactory().get('/')})
    assert urllib.parse.parse_qs(urllib.parse.urlparse(html.unescape(t.render(ctx))).query) == {
        'name': ['Ayers']
    }
    ctx = Context({'request': RequestFactory().get('/?gender=male')})
    assert urllib.parse.parse_qs(urllib.parse.urlparse(html.unescape(t.render(ctx))).query) == {
        'name': ['Ayers']
    }


def test_with_template():
    class ExampleForm1(forms.Form):
        text = forms.CharField(label='Text', max_length=50)

    request = RequestFactory().get('/')
    t = Template('{{ form|with_template }}')
    ctx = Context({'request': request, 'form': ExampleForm1()})
    rendered = t.render(ctx)
    assert PyQuery(rendered).find('input[type=text]')
    assert not PyQuery(rendered).find('input[type=text]').attr['aria-invalid']
    assert PyQuery(rendered).find('input[type=text]').attr['aria-required']

    ctx = Context({'request': request, 'form': ExampleForm1(data=request.GET)})
    rendered = t.render(ctx)
    assert (
        PyQuery(rendered).find('input[type=text][aria-describedby]').attr['aria-describedby']
        == 'error_id_text'
    )
    assert PyQuery(rendered).find('input[type=text]').attr['aria-invalid']

    class ExampleForm2(forms.Form):
        text = forms.CharField(label='Text', max_length=50, help_text='Help text')

    ctx = Context({'request': request, 'form': ExampleForm2()})
    rendered = t.render(ctx)
    assert (
        PyQuery(rendered).find('input[type=text][aria-describedby]').attr['aria-describedby']
        == 'help_text_id_text'
    )

    ctx = Context({'request': request, 'form': ExampleForm2(data=request.GET)})
    rendered = t.render(ctx)
    assert (
        PyQuery(rendered).find('input[type=text][aria-describedby]').attr['aria-describedby']
        == 'help_text_id_text error_id_text'
    )

    class ExampleForm3(forms.Form):
        text = forms.CharField(label='Text', max_length=50, required=False)

    ctx = Context({'request': request, 'form': ExampleForm3(data=request.GET)})
    rendered = t.render(ctx)
    assert not PyQuery(rendered).find('input[type=text]').attr['aria-required']



#! /usr/bin/env python3
# inkscape wrapper to support command-line parameters for <1.0 and 1.0
# versions.

import subprocess
import sys

inkscape_version = subprocess.check_output('inkscape --version', shell=True)
args = sys.argv[1:]
if b'Inkscape 0' not in inkscape_version:
    # --export-png replaced by --export-filename
    # --without-gui and --file removed
    args = [
        x.replace('--export-png', '--export-filename') for x in args if x not in ('--without-gui', '--file')
    ]

sys.exit(subprocess.call(['inkscape'] + args))



import re
import time
from collections import OrderedDict

from django import template
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.forms import BoundField
from django.template import TemplateSyntaxError
from django.utils.html import escape
from django.utils.http import urlencode

register = template.Library()


@register.simple_tag
def xstatic(modname, filename):
    if settings.DEBUG:
        filename = filename.replace('.min.', '.')
    return settings.STATIC_URL + 'xstatic/' + filename


START_TIMESTAMP = time.strftime('%Y%m%d.%H%M')


@register.simple_tag
def start_timestamp():
    return START_TIMESTAMP


# {% querystring %} bits originally from django-tables2.
kwarg_re = re.compile(r'(?:(.+)=)?(.+)')


def token_kwargs(bits, parser):
    """
    Based on Django's `~django.template.defaulttags.token_kwargs`, but with a
    few changes:

    - No legacy mode.
    - Both keys and values are compiled as a filter
    """
    if not bits:
        return {}
    kwargs = OrderedDict()
    while bits:
        match = kwarg_re.match(bits[0])
        if not match or not match.group(1):
            return kwargs
        key, value = match.groups()
        del bits[:1]
        kwargs[parser.compile_filter(key)] = parser.compile_filter(value)
    return kwargs


class QuerystringNode(template.Node):
    def __init__(self, updates, removals):
        super().__init__()
        self.updates = updates
        self.removals = removals

    def render(self, context):
        if not 'request' in context:
            raise ImproperlyConfigured('Missing django.core.context_processors.request')
        params = dict(context['request'].GET)
        for key, value in self.updates.items():
            key = key.resolve(context)
            value = value.resolve(context)
            if key not in ('', None):
                params[key] = value
        for removal in self.removals:
            params.pop(removal.resolve(context), None)
        return escape('?' + urlencode(params, doseq=True))


@register.tag
def querystring(parser, token):
    """
    Creates a URL (containing only the querystring [including "?"]) derived
    from the current URL's querystring, by updating it with the provided
    keyword arguments.

    Example (imagine URL is ``/abc/?gender=male&name=Brad``)::

        {% querystring "name"="Ayers" "age"=20 %}
        ?name=Ayers&gender=male&age=20
        {% querystring "name"="Ayers" without "gender" %}
        ?name=Ayers

    """
    bits = token.split_contents()
    tag = bits.pop(0)
    updates = token_kwargs(bits, parser)
    # ``bits`` should now be empty of a=b pairs, it should either be empty, or
    # have ``without`` arguments.
    if bits and bits.pop(0) != 'without':
        raise TemplateSyntaxError("Malformed arguments to '%s'" % tag)
    removals = [parser.compile_filter(bit) for bit in bits]
    return QuerystringNode(updates, removals)


@register.filter
def with_template(form):
    form_template = template.loader.get_template('gadjo/form.html')
    fields_with_templates = []
    for field in form:
        widget = field.field.widget
        templates = ['gadjo/widget.html']
        if hasattr(widget, 'input_type'):
            templates.insert(0, 'gadjo/%s-widget.html' % widget.input_type)
        aria_described_by = []
        if field.field.help_text:
            aria_described_by.append(f'help_text_{field.id_for_label}')
        if field.errors:
            aria_described_by.append(f'error_{field.id_for_label}')
            field.field.widget.attrs['aria-invalid'] = 'true'
        if field.field.required:
            field.field.widget.attrs['aria-required'] = 'true'
        if aria_described_by:
            field.field.widget.attrs['aria-describedby'] = ' '.join(aria_described_by)
        fields_with_templates.append(
            (
                field,
                template.loader.select_template(templates),
            )
        )
    return form_template.render({'form': form, 'fields_with_templates': fields_with_templates})


# pattern to transform Django camel case class names to CSS class names with
# dashes. (CheckboxInput -> checkbox-input)
class_name_pattern = re.compile(r'(?<!^)(?=[A-Z])')


@register.filter
def field_class_name(field):
    if isinstance(field, BoundField):
        field = field.field
    return class_name_pattern.sub('-', field.widget.__class__.__name__).lower()


@register.filter(name='gadjo_private_substitute_form_prefix')
def form_prefix(value, form):
    if value.startswith('$prefix'):
        value = value.removeprefix('$prefix')
        if form.prefix:
            value = f'{form.prefix}-{value}'
    return value



#! /usr/bin/env python

import glob
import os
import re
import shutil
import subprocess
import sys
import tempfile
import xml.etree.ElementTree as ET

from setuptools import Command, find_packages, setup
from setuptools.command.build import build as _build
from setuptools.command.install_lib import install_lib as _install_lib
from setuptools.command.sdist import sdist
from setuptools.errors import CompileError

inkscape = os.path.abspath(os.path.join(os.path.dirname(__file__), 'inkscape_wrapper.py'))


class eo_sdist(sdist):
    def run(self):
        print('creating VERSION file')
        if os.path.exists('VERSION'):
            os.remove('VERSION')
        version = get_version()
        version_file = open('VERSION', 'w')
        version_file.write(version)
        version_file.close()
        sdist.run(self)
        print('removing VERSION file')
        if os.path.exists('VERSION'):
            os.remove('VERSION')


def get_version():
    '''Use the VERSION, if absent generates a version with git describe, if not
    tag exists, take 0.0- and add the length of the commit log.
    '''
    if os.path.exists('VERSION'):
        with open('VERSION') as v:
            return v.read()
    if os.path.exists('.git'):
        p = subprocess.Popen(
            ['git', 'describe', '--dirty=.dirty', '--match=v*'],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        result = p.communicate()[0]
        if p.returncode == 0:
            result = result.decode('ascii').strip()[1:]  # strip spaces/newlines and initial v
            if '-' in result:  # not a tagged version
                real_number, commit_count, commit_hash = result.split('-', 2)
                version = '%s.post%s+%s' % (real_number, commit_count, commit_hash)
            else:
                version = result.replace('.dirty', '+dirty')
            return version
        else:
            return '0.0.post%s' % len(subprocess.check_output(['git', 'rev-list', 'HEAD']).splitlines())
    return '0.0'


class compile_translations(Command):
    description = 'compile message catalogs to MO files via django compilemessages'
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        curdir = os.getcwd()
        try:
            from django.core.management import call_command

            for path, dirs, files in os.walk('gadjo'):
                if 'locale' not in dirs:
                    continue
                os.chdir(os.path.realpath(path))
                call_command('compilemessages')
                os.chdir(curdir)
        except ImportError:
            sys.stderr.write('!!! Please install Django >= 1.4 to build translations\n')
        finally:
            os.chdir(curdir)


class compile_scss(Command):
    description = 'compile scss files into css files'
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        sass_bin = None
        for program in ('sassc', 'sass'):
            sass_bin = shutil.which(program)
            if sass_bin:
                break
        if not sass_bin:
            raise CompileError(
                'A sass compiler is required but none was found.  See sass-lang.com for choices.'
            )

        for package in self.distribution.packages:
            for package_path in __import__(package).__path__:
                for path, dirnames, filenames in os.walk(package_path):
                    for filename in filenames:
                        if not filename.endswith('.scss'):
                            continue
                        if filename.startswith('_'):
                            continue
                        subprocess.check_call(
                            [
                                sass_bin,
                                '%s/%s' % (path, filename),
                                '%s/%s' % (path, filename.replace('.scss', '.css')),
                            ],
                            env={'LC_ALL': 'C.UTF-8'},
                        )


class build_icons(Command):
    description = 'build icons'
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        destpath = 'gadjo/static/css/icons/'
        if not os.path.exists(destpath):
            os.mkdir(destpath)
        variants_applications = {
            'small': {'colour': '386ede', 'width': '40'},
            'large': {'colour': 'e7e7e7', 'width': '80'},
            'large-hover': {'colour': 'bebebe', 'width': '80'},
        }
        variants_actions = {
            'small': {'colour': '386ede', 'width': '40'},
            'small.white': {'colour': 'ffffff', 'width': '40'},
            'hover': {'colour': '2b2b2b', 'width': '40'},
        }
        for basepath, dirnames, filenames in os.walk('icons'):
            for filename in filenames:
                basename = os.path.splitext(filename)[0]
                variants = variants_applications
                if not filename.endswith('.svg'):
                    continue
                if filename.startswith('action-'):
                    variants = variants_actions
                for variant in variants:
                    dest_filename = '%s.%s.png' % (basename, variant)
                    destname = os.path.join(destpath, dest_filename)
                    self.generate(os.path.join(basepath, filename), destname, **variants.get(variant))

    def generate(self, src, dest, colour, width, **kwargs):
        if os.path.exists(dest) and os.stat(dest).st_mtime >= os.stat(src).st_mtime:
            return
        # default values
        from PIL import Image, PngImagePlugin

        license = 'Creative Commons Attribution-Share Alike 3.0'
        if 'old-set' in src:
            author = 'GNOME Project'
            tree = ET.fromstring(open(src).read().replace('#000000', '#%s' % colour))
            for elem in tree.findall('*'):
                if not elem.attrib.get('style'):
                    elem.attrib['style'] = 'fill:#%s' % colour
        else:
            author = "J'articule"
            tree = ET.fromstring(open(src).read())
            for elem in tree.findall('{http://www.w3.org/2000/svg}defs/{http://www.w3.org/2000/svg}style'):
                elem.text = elem.text.replace('242d3c', colour)

        f = tempfile.NamedTemporaryFile(suffix='.svg', delete=False)
        f.write(ET.tostring(tree))
        f.close()

        subprocess.call(
            [
                inkscape,
                '--without-gui',
                '--file',
                f.name,
                '--export-area-drawing',
                '--export-area-snap',
                '--export-png',
                dest,
                '--export-width',
                width,
            ]
        )

        # write down licensing info in the png file
        meta = PngImagePlugin.PngInfo()
        meta.add_text('Licence', license, 0)
        png_file = Image.open(dest)
        png_file.save(dest, 'PNG', pnginfo=meta)


class build(_build):
    sub_commands = [
        ('compile_translations', None),
        ('compile_scss', None),
        ('build_icons', None),
    ] + _build.sub_commands


class install_lib(_install_lib):
    def run(self):
        self.run_command('compile_translations')
        _install_lib.run(self)


setup(
    name='gadjo',
    version=get_version(),
    license='AGPLv3 or later',
    description='Django base template tailored for management interfaces',
    url='https://dev.entrouvert.org/projects/gadjo/',
    author='Frederic Peters',
    author_email='fpeters@entrouvert.com',
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        'XStatic>=1.0.3',
        'XStatic_Font_Awesome<5',
        'XStatic_jQuery',
        'XStatic_jquery_ui',
        'XStatic_OpenSans',
    ],
    setup_requires=[
        'Pillow',
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU Affero General Public License v3 or later (AGPLv3+)',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
    ],
    zip_safe=False,
    cmdclass={
        'build': build,
        'build_icons': build_icons,
        'compile_scss': compile_scss,
        'compile_translations': compile_translations,
        'install_lib': install_lib,
        'sdist': eo_sdist,
    },
)



import shlex
from pathlib import Path

import nox


def run_hook(name, *args, **kwargs):
    for file in [Path(__name__).parent / '.nox-hooks.py', Path('~/.config/nox/eo-hooks.py').expanduser()]:
        if not file.exists():
            continue

        globals_ = {}
        exec(file.read_text(), globals_)
        hook = globals_.get(name, None)
        if hook:
            hook(*args, **kwargs)


def setup_venv(session, *packages, django_version='>=4.2,<4.3'):
    packages = [
        f'django{django_version}',
        'pytest',
        'pyquery',
        *packages,
    ]
    run_hook('setup_venv', session, packages)
    session.install('-e', '.', *packages, silent=False)


def hookable_run(session, *args, **kwargs):
    args = list(args)
    run_hook('run', session, args, kwargs)
    session.run(*args, **kwargs)


@nox.session()
@nox.parametrize('django', ['>=4.2,<4.3'])
def tests(session, django):
    setup_venv(
        session,
        'pytest-cov',
        'pytest-django',
        'psycopg2-binary',
        'psycopg2',
        django_version=django,
    )

    args = ['py.test']
    if '--coverage' in session.posargs or not session.interactive:
        while '--coverage' in session.posargs:
            session.posargs.remove('--coverage')
        args += [
            '-v',
            '--cov=gadjo/',
            '--cov=tests/',
            '--cov-report=term-missing',
            '--cov-config=.coveragerc',
        ]
        if not session.interactive:
            args += [
                '--cov-report=xml',
                '--cov-report=html',
                f'--junitxml=junit-coverage.django-{django}.xml',
            ]

    if not session.interactive:
        args += ['-v']

    args += session.posargs + ['tests/']

    hookable_run(
        session,
        *args,
    )


@nox.session
def pylint(session):
    setup_venv(session, 'pylint', 'astroid<3', 'pylint-django', 'nox')
    pylint_command = ['pylint', '--jobs', '12', '-f', 'parseable', '--rcfile', 'pylint.rc']

    if not session.posargs:
        pylint_command += ['gadjo/', 'tests/', 'noxfile.py']
    else:
        pylint_command += session.posargs

    if not session.interactive:
        session.run(
            'bash',
            '-c',
            f'{shlex.join(pylint_command)} | tee pylint.out ; test $PIPESTATUS -eq 0',
            external=True,
        )
    else:
        session.run(*pylint_command)


@nox.session
def codestyle(session):
    session.install('pre-commit')
    session.run('pre-commit', 'run', '--all-files', '--show-diff-on-failure')


@nox.session
def check_manifest(session):
    # django is only required to compile messages
    session.install('django', 'check-manifest')
    # compile messages and css
    ignores = [
        'VERSION',
        'gadjo/static/css/*.css',
        'gadjo/static/css/icons/*.png',
        'merge-junit-results.py',
    ]
    session.run('check-manifest', '--ignore', ','.join(ignores))
